/********************************************************************/
/*    TORQUE (Types Of Relationships QUality Evaluation) toolkit    */
/*                                                                  */
/* Original file name: fileparsers.cc                               */
/*                                                                  */
/* This software by R.M.                                            */
/********************************************************************/

#include "fileparsers.h"

// ******************************************
// ************ BGP dump parser *************
// ******************************************

/* Here goes the initialization of the header names of each field in the BGP
   dump. If ever a field header or status character had changed, just
   update its entry right here; warning: header names must be in the same
   order as they appear in the above corresponding enum types */
const char bgpParser::fieldCaptions[7][9] =
   {"Network", "Next Hop", "Metric", "LocPrf", "Weight", "Path", "--More--"};
const char bgpParser::routeStatCodes[7] =
   {'s', 'd', 'h', '*', '>', 'i', 'S'};
const char bgpParser::routeOriginCodes[3] =
   {'i', 'e', '?'};

// And here goes the initialization of the special delimiters
const char bgpParser::privateAS[2] = {'(', ')'};
const char bgpParser::ASset[2] = {'{', '}'};


// Default constructor
bgpParser::bgpParser() {
   parsedData.netmask = NULL;
   parsedData.metric = NULL;
   parsedData.locpref = NULL;
   parsedData.weight = NULL;
   lastNetwork = NULL;
   lastNetmask = NULL;
}

// Destructor
bgpParser::~bgpParser() {
   if (lastNetwork != NULL) {
      delete lastNetwork;
   }
   if (lastNetmask != NULL) {
      delete lastNetmask;
   }
   if (parsedData.netmask != NULL) {
      delete parsedData.netmask;
   }
   if (parsedData.metric != NULL) {
      delete parsedData.metric;
   }
   if (parsedData.locpref != NULL) {
      delete parsedData.locpref;
   }
   if (parsedData.weight != NULL) {
      delete parsedData.weight;
   }
}

/* Attempts to parse current line; before doing that, you must use nextLine in
   order to move over the first parsable line; if you forget doing so, a
   PARSE_NOLINE error will be returned */
parse_stat bgpParser::parseLine() {
   int i, prefixLen, offset, netmask;
   unsigned int curLineLen = strlen(curLine);
   long int value;
   unsigned long long int value2;
   int bytequartet[4];
   char *curToken, *err;
   bool foundLine, interpreted, privas, asset, dontGoOn;
   ASItem asitem;

   strcpy(error, "");
   if (parseFile == NULL) {
      strcat (error, "No file has been opened");
      return (PARSE_NOFILE);
   }
   if (strlen(curLine) == 0) {
      strcat (error, "There is no current line to be parsed");
      return (PARSE_NOLINE);
   }
   // Parse current line
   strcpy(copiedLine, curLine);

   // ================================
   // Route attributes
   // ================================
   parsedData.route_flags = 0;
   for (i=0; i<=2; i++) {
      interpreted = false;
      if (copiedLine[i] == routeStatCodes[0]) {
         parsedData.route_flags |= ROUTE_SUPPRESSED;
         interpreted = true;
      }
      if (copiedLine[i] == routeStatCodes[1]) {
         parsedData.route_flags |= ROUTE_DAMPED;
         interpreted = true;
      }
      if (copiedLine[i] == routeStatCodes[2]) {
         parsedData.route_flags |= ROUTE_HISTORY;
         interpreted = true;
      }
      if (copiedLine[i] == routeStatCodes[3]) {
         parsedData.route_flags |= ROUTE_VALID;
         interpreted = true;
      }
      if (copiedLine[i] == routeStatCodes[4]) {
         parsedData.route_flags |= ROUTE_BEST;
         interpreted = true;
      }
      if (copiedLine[i] == routeStatCodes[5]) {
         parsedData.route_flags |= ROUTE_INTERNAL;
         interpreted = true;
      }
      if (copiedLine[i] == routeStatCodes[6]) {
         parsedData.route_flags |= ROUTE_STALE;
         interpreted = true;
      }
      if (!interpreted && copiedLine[i]!=' ') {
         strcat (error, "Unrecognized route status character: `");
         // Now append a single character to the error string
         value = strlen(error);
         error[value] = copiedLine[i];
         error[value+1] = '\0';
         strcat (error, "'");
         return (PARSE_ILLEGAL_VALUE);
      }
   }

   // ================================
   // Network
   // ================================
   if (fieldPos[FIELD_NETWORK] == 0) {
      strcat (error, "Could not determine position of network field");
      return (PARSE_MISMATCH);
   }
   prefixLen = 0;
   if (curLineLen >= (unsigned int) fieldPos[FIELD_NETWORK] &&
       curLine[fieldPos[FIELD_NETWORK]-1]!=' ' &&
       curLine[fieldPos[FIELD_NETWORK]-1]!='\t') {
      curToken = strtok(copiedLine + fieldPos[FIELD_NETWORK] - 1, " \t");
      // Store prefix length (characters) for later use
      prefixLen = strlen(curToken);
      netmask = -1;
      if (curToken == NULL) {
         strcat (error, "Invalid network");
         return (PARSE_MISMATCH);
      }
      if (strchr(curToken, '/') != NULL) {
         // There are both IP address and netmask
         if (sscanf(curToken, "%i.%i.%i.%i/%i ", bytequartet, bytequartet + 1,
                                                 bytequartet + 2, bytequartet + 3,
                                                 &netmask) < 5) {
            strcat (error, "Invalid network address/netmask: `");
            strcat (error, curToken);
            strcat (error, "'");
            return (PARSE_ILLEGAL_VALUE);
         }
         if (netmask < 0 || netmask > 32) {
            strcat (error, "Invalid netmask value: `");
            sprintf(error, "%s%i'", error, netmask);
            return (PARSE_ILLEGAL_VALUE);
         }
      }
      else {
         // Only IP address
         if (sscanf(curToken, "%i.%i.%i.%i ", bytequartet, bytequartet + 1,
                                              bytequartet + 2, bytequartet + 3) < 4) {
            strcat (error, "Invalid network address: `");
            strcat (error, curToken);
            strcat (error, "'");
            return (PARSE_ILLEGAL_VALUE);
         }
      }
      for (i=0; i<=3; i++) {
         if (bytequartet[i] < 0 || bytequartet[i] > 255) {
            strcat (error, "Invalid network byte value: `");
            sprintf(error, "%s%i'", error, bytequartet[i]);
            return (PARSE_ILLEGAL_VALUE);
         }
      }
      // There is a prefix
      if (lastNetwork == NULL) {
         lastNetwork = new IPAddress;
      }
      *lastNetwork =
         IPAddress(bytequartet[0], bytequartet[1], bytequartet[2], bytequartet[3]);
      if (netmask > 0) {
         if (lastNetmask == NULL) {
            lastNetmask = new IPNetmask;
         }
         *lastNetmask = IPNetmask(netmask);
      }
      else {
         if (lastNetmask != NULL) {
            delete lastNetmask;
            lastNetmask = NULL;
         }
      }
   }
   if (lastNetwork == NULL) {
      /* Current line contains no network field, and there are no
         previous networks */
      strcat (error, "Could not retrieve any previous network value");
      return (PARSE_MISMATCH);
   }
   parsedData.network = *lastNetwork;
   if (lastNetmask == NULL) {
      if (parsedData.netmask != NULL) {
         delete parsedData.netmask;
         parsedData.netmask = NULL;
      }
   }
   else {
      if (parsedData.netmask == NULL) {
         parsedData.netmask = new IPNetmask;
      }
      *(parsedData.netmask) = *lastNetmask;
   }

   // ================================
   // Next hop
   // ================================
   if (fieldPos[FIELD_NEXTHOP] == 0) {
      strcat (error, "Could not determine position of next hop field");
      return (PARSE_MISMATCH);
   }
   if (prefixLen > 16) {
      /* In some routing tables, when the network is longer than 16 characters,
         the next hop field is either moved rightward or put on a new line */
      offset = prefixLen - 16;
   }
   else {
      offset = 0;
   }
   if (curLineLen < (unsigned int) fieldPos[FIELD_NEXTHOP] + offset) {
      // Line is too short: check if next hop is on next line
      foundLine = false;
      offset = 0;
      while (!foundLine && !ferror(parseFile) && !feof(parseFile)) {
         parsedlines++;                                     // Parsing of current line is over
         if (fgets(curLine, 513, parseFile) == NULL) {      // 513 = 512 + `\0'
            // End of file reached
            strcpy (curLine, "");
         }
         readlines++;
         curLineLen = strlen(curLine);       // Do not consider newline character
         parseFilePos += curLineLen;
         if (strlen(curLine) > 3 && strstr(curLine, fieldCaptions[6]) == NULL) {
            // This is neither a "short" nor a --More-- line
            foundLine = true;
            // Remove \n and/or \r at end of line
            value = strlen(curLine) - 1;
            if (curLine[value-1] == '\n' || curLine[value-1] == '\r') {
               curLine[value-1] = '\0';
               curLineLen -= 2;
            }
            else {
               if (curLine[value] == '\n' || curLine[value] == '\r') {
                  curLine[value] = '\0';
                  curLineLen--;
               }
            }
         }
      }
      if (!foundLine) {
         strcat (error, "Unexpected end of file (next hop expected)");
         return (PARSE_UNEXPECTED_EOF);
      }
      strcpy (copiedLine, curLine);
   }
   curToken = strtok(copiedLine + fieldPos[FIELD_NEXTHOP] + offset - 1, " \t");
   if (curToken == NULL) {
      strcat (error, "Invalid next hop");
      return (PARSE_MISMATCH);
   }
   if (sscanf(curToken, "%i.%i.%i.%i ", bytequartet, bytequartet + 1,
                                       bytequartet + 2, bytequartet + 3) < 4) {
      strcat (error, "Invalid next hop: `");
      strcat (error, curToken);
      strcat (error, "'");
      return (PARSE_ILLEGAL_VALUE);
   }
   for (i=0; i<=3; i++) {
      if (bytequartet[i] < 0 || bytequartet[i] > 255) {
         strcat (error, "Invalid next hop byte value: `");
         sprintf(error, "%s%i'", error, bytequartet[i]);
         return (PARSE_ILLEGAL_VALUE);
      }
   }
   parsedData.nextHop = IPAddress(
      bytequartet[0], bytequartet[1], bytequartet[2], bytequartet[3]);

   // ================================
   // Metric
   // ================================
   if (fieldPos[FIELD_METRIC] > 0) {
      // There can be a metric field
      if (curLineLen >= (unsigned int) fieldPos[FIELD_METRIC]) {
         i = fieldPos[FIELD_METRIC] + offset + strlen(fieldCaptions[FIELD_METRIC]) - 2;
         if (copiedLine[i] != ' ' &&
             copiedLine[i] != '\t') {
            i = fieldPos[FIELD_METRIC] + offset - 1;
            /* Some metric values can be longer than the "Metric" word: find
               metric field beginning by moving backwards */
            while (i > 0 && copiedLine[i] != ' ') {
               i--;
            }
            curToken = strtok (copiedLine + i, " \t");
            if (curToken != NULL) {
               value2 = strtoull(curToken, &err, 10);
               if (strlen(err) > 0) {
                  strcat (error, "Invalid metric value: `");
                  strcat (error, curToken);
                  strcat (error, "'");
                  return (PARSE_ILLEGAL_VALUE);
               }
               if (parsedData.metric == NULL) {
                  parsedData.metric = new unsigned long long int;
               }
               *(parsedData.metric) = value2;
            }
         }
      }
   }
   else {
      // Could not determine the position of the metric field
      if (parsedData.metric != NULL) {
         delete parsedData.metric;
      }
      parsedData.metric = NULL;
   }

   // ================================
   // Local preference
   // ================================
   if (fieldPos[FIELD_LOCPRF] > 0) {
      // There can be a LocPrf field
      if (curLineLen >= (unsigned int) fieldPos[FIELD_LOCPRF]) {
         i = fieldPos[FIELD_LOCPRF] + offset + strlen(fieldCaptions[FIELD_LOCPRF]) - 2;
         if (copiedLine[i] != ' ' &&
             copiedLine[i] != '\t') {
            curToken = strtok (copiedLine + fieldPos[FIELD_LOCPRF] + offset - 1, " \t");
            if (curToken != NULL) {
               value = strtol(curToken, &err, 10);
               if (value < 0 || value > 65535 || (strlen(err) > 0)) {
                  strcat (error, "Invalid local preference value: `");
                  strcat (error, curToken);
                  strcat (error, "'");
                  return (PARSE_ILLEGAL_VALUE);
               }
               if (parsedData.locpref == NULL) {
                  parsedData.locpref = new unsigned int;
               }
               *(parsedData.locpref) = value;
            }
         }
      }
   }
   else {
      // Could not determine the position of the local preference field
      if (parsedData.locpref != NULL) {
         delete parsedData.locpref;
      }
      parsedData.locpref = NULL;
   }

   // ================================
   // Weight
   // ================================
   if (fieldPos[FIELD_WEIGHT] > 0) {
      // There can be a weight field
      if (curLineLen >= (unsigned int) fieldPos[FIELD_WEIGHT]) {
         i = fieldPos[FIELD_WEIGHT] + offset + strlen(fieldCaptions[FIELD_WEIGHT]) - 2;
         if (copiedLine[i] != ' ' &&
             copiedLine[i] != '\t') {
            curToken = strtok (copiedLine + fieldPos[FIELD_WEIGHT] + offset - 1, " \t");
            if (curToken != NULL) {
               value = strtol(curToken, &err, 10);
               if (value < 0 || value > 65535 || (strlen(err) > 0)) {
                  strcat (error, "Invalid weight value: `");
                  strcat (error, curToken);
                  strcat (error, "'");
                  return (PARSE_ILLEGAL_VALUE);
               }
               if (parsedData.weight == NULL) {
                  parsedData.weight = new unsigned int;
               }
               *(parsedData.weight) = value;
            }
         }
      }
   }
   else {
      // Could not determine the position of the weight field
      if (parsedData.weight != NULL) {
         delete parsedData.weight;
      }
      parsedData.weight = NULL;
   }

   // ================================
   // AS path
   // ================================
   if (fieldPos[FIELD_ASPATH] == 0) {
      strcat (error, "Could not determine position of the AS path field");
      return (PARSE_MISMATCH);
   }
   if (curLineLen < (unsigned int) fieldPos[FIELD_ASPATH] + offset) {
      strcat (error, "No AS path");
      return (PARSE_MISMATCH);
   }
   curToken = strtok(copiedLine + fieldPos[FIELD_ASPATH] + offset - 1, " ,");
   // Empty AS path list
   parsedData.ASpath = List<ASItem>();
   parsedData.path_flags = 0;
   privas = false;
   asset = false;
   while (curToken != NULL) {
      dontGoOn = false;
      asitem.flags = 0;
      if (privas || (curToken[0] == privateAS[0] && curToken[1] != '\0')) {
         asitem.flags |= AS_PRIVATE;
         if (!privas) {
            // There is a `(' character to be skipped
            curToken++;
         }
         if (privas && (curToken[0] == privateAS[0] && curToken[1] != '\0')) {
            strcat (error, "Parentheses mismatch in AS path");
            return (PARSE_MISMATCH);
         }
         privas = true;
         if (asset) {
            strcat (error, "Nested AS attributes");
            return (PARSE_ILLEGAL_VALUE);
         }
      }
      if (asset || (curToken[0] == ASset[0] && curToken[1] != '\0')) {
         asitem.flags |= AS_SET;
         if (!asset) {
            // There is a `{' character to be skipped
            curToken++;
         }
         if (asset && (curToken[0] == ASset[0] && curToken[1] != '\0')) {
            strcat (error, "Parentheses mismatch in AS path");
            return (PARSE_MISMATCH);
         }
         asset = true;
         if (privas) {
            strcat (error, "Nested AS attributes");
            return (PARSE_ILLEGAL_VALUE);
         }
      }
      /* If either of the three following conditions holds, then
         current token is not an AS number but a route
         origin flag: parse it and skip to next token */
      if (curToken[0] == routeOriginCodes[0] && curToken[1] == '\0') {
         parsedData.path_flags |= ORIGIN_IGP;
         dontGoOn = true;
      }
      if (curToken[0] == routeOriginCodes[1] && curToken[1] == '\0') {
         parsedData.path_flags |= ORIGIN_EGP;
         dontGoOn = true;
      }
      if (curToken[0] == routeOriginCodes[2] && curToken[1] == '\0') {
         parsedData.path_flags |= ORIGIN_INCOMPLETE;
         dontGoOn = true;
      }
      if (!dontGoOn && strlen(curToken) == 0) {
         strcat (error, "Invalid AS number or path flag: `");
         strcat (error, curToken);
         strcat (error, "'");
         return (PARSE_ILLEGAL_VALUE);
      }
      i = strlen(curToken) - 1;
      if (curToken[i] == privateAS[1] && i > 0) {
         /* Overwrite last parenthesis in order to
            be ready to parse AS number later on */
         if (!privas) {
            strcat (error, "Parentheses mismatch in AS path");
            return (PARSE_MISMATCH);
         }
         curToken[i] = '\0';
         privas = false;
      }
      if (curToken[i] == ASset[1] && i > 0) {
         /* Overwrite last parenthesis in order to
            be ready to parse AS number later on */
         if (!asset) {
            strcat (error, "Parentheses mismatch in AS path");
            return (PARSE_MISMATCH);
         }
         curToken[i] = '\0';
         asset = false;
      }
      if (!dontGoOn) {
         value = strtol(curToken, &err, 10);
         if (value < 1 || value > 65535 || (strlen(err) > 0)) {
            strcat (error, "Invalid AS number or path flag: `");
            strcat (error, curToken);
            strcat (error, "'");
            return (PARSE_ILLEGAL_VALUE);
         }
         if (value >= 64512) {
            // This is a private AS, even if not explicitly declared
            asitem.flags |= AS_PRIVATE;
         }
         asitem.ASnum = value;
         parsedData.ASpath.enqueue(asitem);
      }
      curToken = strtok(NULL, " ,");
   }
   if (privas || asset) {
      strcat (error, "Parentheses mismatch in AS path");
      return (PARSE_MISMATCH);
   }

   parsedlines++;
   return (PARSE_NOERROR);
}

// Moves to the next parsable line in the file
parsefile_stat bgpParser::nextLine() {
   bool foundParsableLine, foundSpecialChar;
   int i, i2;
   char *strPos, *curToken;
   char flags;
   strcpy (error, "");

   if (parseFile == NULL) {
      strcpy (error, "No file has been opened");
      return (PARSEFILE_NOFILE);
   }
   if (feof(parseFile)) {
      return (PARSEFILE_EOF);
   }

   if (!nowParsing) {
      /* If `nowParsing' is false, this is a comment area, and
         so there cannot be any previous network prefix */
      if (lastNetwork != NULL) {
         delete lastNetwork;
      }
      lastNetwork = NULL;
   }

   foundParsableLine = false;
   while (!foundParsableLine && !feof(parseFile) && (!ferror(parseFile))) {
      if (fgets(curLine, 513, parseFile) == NULL) {        // 513 = 512 + `\0'
         // End of file reached
         strcpy (curLine, "");
      }
      else {
         readlines++;
      }
      parseFilePos += strlen(curLine);
      if (strlen(curLine) > 3) {
         if (!ferror(parseFile)) {
            /* Since strtok changes string, it is better to work
               on a copy of `curLine' */
            strcpy(copiedLine, curLine);
            // Now test whether the first word in the line is `Network'
            curToken = strtok(copiedLine, " \t");
            strPos = strstr(curToken, fieldCaptions[FIELD_NETWORK]);
            if (strPos != NULL) {
               /* This is the absolutely first header line of the BGP dump:
                  save field positions (columns) for later use */
               for (i=0; i<=6; i++) {
                  strPos = strstr(curLine, fieldCaptions[i]);
                  if (strPos != NULL) {
                     fieldPos[i] = strPos - curLine + 1;
                  }
               }
               // Parsing should now start normally
               nowParsing = true;
            }
            else {
               if (nowParsing) {
                  /* Test whether this is an intermediate header
                     row (i.e., one starting with `--More--') */
                  strPos = strstr(curToken, fieldCaptions[FIELD_MORE]);
                  if (strPos == NULL) {
                     // This is either a comment line or a parsable line
                     foundParsableLine = true;
                     flags = 0;
                     /* Examine the first 3 characters of the line to
                        determine whether it is parsable */
                     for (i2=0; i2<=2 && foundParsableLine; i2++) {
                        foundSpecialChar = false;
                        for (i=0; i<=6 && !foundSpecialChar; i++) {
                           if (curLine[i2]==routeStatCodes[i]) {
                              // Test if there is duplicate use of a route flag
                              foundSpecialChar = !(flags & (1 << (i+1)));
                              flags |= 1 << (i+1);
                           }
                           if (curLine[i2]==' ') {
                              foundSpecialChar = true;
                           }
                        }
                        foundParsableLine = foundParsableLine && foundSpecialChar;
                        nowParsing = nowParsing && foundParsableLine;
                     }
                  }
               }
               else {
                  /* Intermediate header lines can mark the beginning of
                     a parsable zone; this usually happens when the first
                     three characters of a line do not match expected route
                     attributes, but that is indeed a parsable line; in such
                     a case, if intermediate header lines were not considered,
                     the rest of the file would be discarded */
                  strPos = strstr(curToken, fieldCaptions[FIELD_MORE]);
                  if (strPos != NULL) {
                     nowParsing = true;
                  }
               }
            }
         }
      }
   }
   if (!foundParsableLine) {
      strcpy(curLine, "");
      return (PARSEFILE_EOF);
   }
   else {
      // Replace \n and \r at end of line with \0
      if (strpbrk(curLine, "\n\r")) {
         *(strpbrk(curLine, "\n\r")) = '\0';
      }
   }
   return (PARSEFILE_NOERROR);
}

// ************ BGP dump parser *************






// ******************************************
// ********* bgpSat output parser ***********
// ******************************************

/* Change the following line if you need to deal with
   peering/bad path lines that are identified by
   different strings */
const char satParser::peerStr[18] = "candidate peering";
const char satParser::badStr[10] = "bad path:";

// Default constructor
satParser::satParser() {
   parsedData.straight = NULL;
   parsedData.negated = NULL;
}

// Destructor
satParser::~satParser() {
   if (parsedData.straight != NULL) {
      delete parsedData.straight;
   }
   if (parsedData.negated != NULL) {
      delete parsedData.negated;
   }
}

/* Attempts to parse current line; before doing that, you must use nextLine in
   order to move over the first parsable line; if you forget doing so, a
   PARSE_NOLINE error will be returned */
parse_stat satParser::parseLine() {
   char copiedLine[513];
   char *curToken, *err;
   long int value;
   double value2;
   int i;

   strcpy (error, "");
   if (parseFile == NULL) {
      strcat (error, "No file has been opened");
      return (PARSE_NOFILE);
   }
   if (strlen(curLine) == 0) {
      strcat (error, "There is no current line to be parsed");
      return (PARSE_NOLINE);
   }
   parsedData.peering = false;
   parsedData.bad = false;

   /* String operations can overwrite curLine, while
      this should not happen */
   strcpy (copiedLine, curLine);
   curToken = strtok (copiedLine, "\t ");
   if (curToken == NULL) {
      strcat (error, "Line appears to be empty");
      return (PARSE_MISMATCH);
   }
   if (strstr(curLine, peerStr) != NULL) {
      // This is a `candidate peering' line
      parsedData.peering = true;
   }
   if (strstr(curLine, badStr) != NULL) {
      // This is a `bad path' line
      parsedData.bad = true;
   }
   if (parsedData.peering || parsedData.bad) {
      curToken = strtok(NULL, "\t ");
      curToken = strtok(NULL, "\t ");
   }
   // Empty AS path
   parsedData.ASpath = List<unsigned short int>();
   if (!parsedData.bad) {
      // This line holds information about edge direction
      if (curToken == NULL) {
         strcat (error, "Unable to retrieve `from' AS number");
         return (PARSE_MISMATCH);
      }
      value = strtol (curToken, &err, 10);
      if (value < 1 || value > 65535 || (strlen(err) > 0)) {
         strcat (error, "Invalid AS number: `");
         strcat (error, curToken);
         strcat (error, "'");
         return (PARSE_ILLEGAL_VALUE);
      }
      parsedData.from = value;

      // Next instructions captures `->' or `-'
      curToken = strtok (NULL, "\t ");

      curToken = strtok (NULL, "\t ");
      if (curToken == NULL) {
         strcat (error, "Unable to retrieve `to' AS number");
         return (PARSE_MISMATCH);
      }
      value = strtol (curToken, &err, 10);
      if (value < 1 || value > 65535 || (strlen(err) > 0)) {
         strcat (error, "Invalid AS number: `");
         strcat (error, curToken);
         strcat (error, "'");
         return (PARSE_ILLEGAL_VALUE);
      }
      parsedData.to = value;

      if (!parsedData.peering) {
         // Now skip  ASX -> ASY fields
         for (i=1; i<=4; i++) {
            curToken = strtok(NULL, "\t ");
            if (curToken == NULL) {
               strcat (error, "Unable to retrieve reversed flag");
               return (PARSE_MISMATCH);
            }
         }
         value = strtol (curToken, &err, 10);
         if (value < 0 || value > 1 || (strlen(err) > 0)) {
            strcat (err, "Invalid reversed flag: `");
            strcat (err, curToken);
            strcat (err, "'");
            return (PARSE_ILLEGAL_VALUE);
         }
         parsedData.reversed = (value == 0);
         // Move over covering properties
         curToken = strtok(NULL, "\t ");
      }
      else {
         /* Just go to the next field and check if
            there are edge covering properties */
         curToken = strtok(NULL, "\t ");
      }
      if (curToken != NULL) {
         value2 = strtod(curToken, &err);
         if (value2 < 0 || (strlen(err) > 0)) {
            strcat (err, "Invalid edge covering value: `");
            strcat (err, curToken);
            strcat (err, "'");
            return (PARSE_ILLEGAL_VALUE);
         }
         if (parsedData.straight == NULL) {
            parsedData.straight = new double;
         }
         *(parsedData.straight) = value2;
         // Next instruction captures `(str.)'
         curToken = strtok(NULL, "\t ");
         curToken = strtok(NULL, "\t ");
         if (curToken == NULL) {
            strcat (error, "Unable to retrieve edge covering properties");
            return (PARSE_MISMATCH);
         }
         value2 = strtod(curToken, &err);
         if (value2 < 0 || (strlen(err) > 0)) {
            strcat (err, "Invalid edge covering value: `");
            strcat (err, curToken);
            strcat (err, "'");
            return (PARSE_ILLEGAL_VALUE);
         }
         if (parsedData.negated == NULL) {
            parsedData.negated = new double;
         }
         *(parsedData.negated) = value2;
      }
      else {
         // No edge covering properties
         if (parsedData.straight != NULL) {
            delete parsedData.straight;
            parsedData.straight = NULL;
         }
         if (parsedData.negated != NULL) {
            delete parsedData.negated;
            parsedData.negated = NULL;
         }
      }
   }
   else {
      // This is a bad path line, so information is a path
      while (curToken != NULL) {
         value = strtol(curToken, &err, 10);
         if (value < 1 || value > 65535 || (strlen(err) > 0)) {
            strcat (error, "Invalid AS number in bad AS path: `");
            strcat (error, curToken);
            strcat (error, "'");
            return (PARSE_ILLEGAL_VALUE);
         }
         parsedData.ASpath.enqueue(value);
         curToken = strtok (NULL, "\t ");
      }
   }

   return (PARSE_NOERROR);
}

// Moves to the next parsable line in the file
parsefile_stat satParser::nextLine() {
   bool foundParsableLine;
   strcpy (error, "");
   if (parseFile == NULL) {
      strcat (error, "No file has been opened");
      return (PARSEFILE_NOFILE);
   }
   if (feof(parseFile)) {
      return (PARSEFILE_EOF);
   }
   foundParsableLine = false;
   while (!foundParsableLine && !ferror(parseFile) && !feof(parseFile)) {
      if (fgets (curLine, 513, parseFile) == NULL) {
         // End of file reached
         strcpy (curLine, "");
      }
      else {
         readlines++;
      }
      parseFilePos += strlen(curLine);
      if (strlen(curLine) > 5) {
         // This appears to be a valid line
         foundParsableLine = true;
      }
   }
   if (!foundParsableLine) {
      strcpy (curLine, "");
      return (PARSEFILE_EOF);
   }

   // Remove \n or \r at end of line
   if (strpbrk(curLine, "\n\r")) {
      *(strpbrk(curLine, "\n\r")) = '\0';
   }

   return (PARSEFILE_NOERROR);
}

// ********* bgpSat output parser ***********






// ******************************************
// ********* Agarwal graph parser ***********
// ******************************************

/* Attempts to parse current line; before doing that, you must use nextLine in
   order to move over the first parsable line; if you forget doing so, a
   PARSE_NOLINE error will be returned */
parse_stat vantageParser::parseLine() {
   int node1, node2, dir;

   strcpy (error, "");
   if (parseFile == NULL) {
      strcat (error, "No file has been opened");
      return (PARSE_NOFILE);
   }
   if (strlen(curLine) == 0) {
      strcat (error, "There is no current line to be parsed");
      return (PARSE_NOLINE);
   }
   parsedData.peering = false;

   if (sscanf(curLine, "%i %i %i", &node1, &node2, &dir) < 3) {
      strcat (error, "Unable to find either AS number and/or direction");
      return (PARSE_MISMATCH);
   }
   if (node1 < 0 || node1 > 65535) {
      strcat (error, "Invalid AS number: `");
      sprintf(error, "%s%i'", error, node1);
      return (PARSE_ILLEGAL_VALUE);
   }
   if (node2 < 0 || node2 > 65535) {
      strcat (error, "Invalid AS number: `");
      sprintf(error, "%s%i'", error, node2);
      return (PARSE_ILLEGAL_VALUE);
   }
   if (dir != 0 && dir != 1) {
      strcat (error, "Edge direction cannot be `");
      sprintf(error, "%s%i'", error, dir);
      return (PARSE_ILLEGAL_VALUE);
   }
   parsedData.from = node2;
   parsedData.to = node1;
   parsedData.peering = (dir == 0);
   return (PARSE_NOERROR);
}

// Moves to the next parsable line in the file
parsefile_stat vantageParser::nextLine() {
   bool foundParsableLine;
   strcpy (error, "");
   if (parseFile == NULL) {
      strcat (error, "No file has been opened");
      return (PARSEFILE_NOFILE);
   }
   if (feof(parseFile)) {
      return (PARSEFILE_EOF);
   }
   foundParsableLine = false;
   while (!foundParsableLine && !ferror(parseFile) && !feof(parseFile)) {
      if (fgets (curLine, 513, parseFile) == NULL) {
         // End of file reached
         strcpy (curLine, "");
      }
      else {
         readlines++;
      }
      parseFilePos += strlen(curLine);
      if (strlen(curLine) > 4) {
         // This appears to be a valid line
         foundParsableLine = true;
      }
   }
   if (!foundParsableLine) {
      strcpy (curLine, "");
      return (PARSEFILE_EOF);
   }
   // Remove \n or \r at end of line
   if (strpbrk(curLine, "\n\r")) {
      *(strpbrk(curLine, "\n\r")) = '\0';
   }

   return (PARSEFILE_NOERROR);
}

// ********* Agarwal graph parser ***********




// ******************************************
// ********** AS Path file parser ***********
// ******************************************

/* Attempts to parse current line; before doing that, you must use nextLine in
   order to move over the first parsable line; if you forget doing so, a
   PARSE_NOLINE error will be returned */
parse_stat pathParser::parseLine() {
   char *curToken, *err;
   char copiedLine[513];
   long int value;
   bool suspendParsing;

   suspendParsing = false;
   strcpy (error, "");
   if (parseFile == NULL) {
      strcat (error, "No file has been opened");
      return (PARSE_NOFILE);
   }
   if (strlen(curLine) == 0) {
      strcat (error, "There is no current line to be parsed");
      return (PARSE_NOLINE);
   }
   /* Current line will be "tokenized", which overwrites some of
      its characters; the following instruction keeps a safe
      copy of it */
   strcpy (copiedLine, curLine);

   // Empty current list of ASs in the path
   parsedData.ASpath = List<unsigned short int>();

   curToken = strtok(copiedLine, " \t");
   while (curToken != NULL) {
      if (curToken[0] == '(') {
         // Avoid parsing AS sets
         suspendParsing = true;
      }
      if (!suspendParsing) {
         value = strtol(curToken, &err, 10);
         if (value < 1 || value > 65535 || (strlen(err) > 0)) {
            strcat (error, "Invalid AS number: `");
            strcat (error, curToken);
            strcat (error, "'");
            return (PARSE_ILLEGAL_VALUE);
         }
         parsedData.ASpath.enqueue(value);
      }
      if (curToken[strlen(curToken) - 1] == ')') {
         suspendParsing = false;
      }
      curToken = strtok(NULL, " \t");
   }
   return (PARSE_NOERROR);
}

// Moves to the next parsable line in the file
parsefile_stat pathParser::nextLine() {
   bool foundParsableLine;
   strcpy (error, "");
   if (parseFile == NULL) {
      strcat (error, "No file has been opened");
      return (PARSEFILE_NOFILE);
   }
   if (feof(parseFile)) {
      return (PARSEFILE_EOF);
   }
   foundParsableLine = false;
   while (!foundParsableLine && !ferror(parseFile) && !feof(parseFile)) {
      if (fgets (curLine, 513, parseFile) == NULL) {
         // End of file reached
         strcpy (curLine, "");
      }
      else {
         readlines++;
      }
      parseFilePos += strlen(curLine);
      if (strlen(curLine) > 0) {
         // This appears to be a valid line
         foundParsableLine = true;
      }
   }
   if (!foundParsableLine) {
      strcpy (curLine, "");
      return (PARSEFILE_EOF);
   }

   // Remove \n or \r at end of line
   if (strpbrk(curLine, "\n\r")) {
      *(strpbrk(curLine, "\n\r")) = '\0';
   }

   return (PARSEFILE_NOERROR);
}

// ********** AS Path file parser ***********






// ******************************************
// ******* Unknown edges file parser ********
// ******************************************


/* Attempts to parse current line; before doing that, you
   must use nextLine in order to move over the first parsable
   line; if you forget doing so, a PARSE_NOLINE error will
   be returned */
parse_stat unknownParser::parseLine() {
   int from, to;

   if (parseFile == NULL) {
      strcat (error, "No file has been opened");
      return (PARSE_NOFILE);
   }
   if (strlen(curLine) == 0) {
      strcat (error, "There is no current line to be parsed");
      return (PARSE_NOLINE);
   }
   if (sscanf(curLine, "%i %i", &from, &to) < 2) {
      strcat (error, "Invalid edge or AS number: `");
      strcat (error, curLine);
      strcat (error, "'");
      return (PARSE_MISMATCH);
   }
   if (from < 0 || from > 65535) {
      strcat (error, "Invalid AS number: `");
      sprintf(error, "%s%i'", error, from);
      return (PARSE_ILLEGAL_VALUE);
   }
   if (to < 0 || to > 65535) {
      strcat (error, "Invalid AS number: `");
      sprintf(error, "%s%i'", error, to);
      return (PARSE_ILLEGAL_VALUE);
   }
   parsedData.from = from;
   parsedData.to = to;

   return (PARSE_NOERROR);
}

// Moves to the next parsable line in the file
parsefile_stat unknownParser::nextLine() {
   bool foundParsableLine;
   strcpy (error, "");
   if (parseFile == NULL) {
      strcat (error, "No file has been opened");
      return (PARSEFILE_NOFILE);
   }
   if (feof(parseFile)) {
      return (PARSEFILE_EOF);
   }
   foundParsableLine = false;
   while (!foundParsableLine && !ferror(parseFile) && !feof(parseFile)) {
      if (fgets (curLine, 513, parseFile) == NULL) {
         // End of file reached
         strcpy (curLine, "");
      }
      else {
         readlines++;
      }
      parseFilePos += strlen(curLine);
      if (strlen(curLine) > 0) {
         // This appears to be a valid line
         foundParsableLine = true;
      }
   }
   if (!foundParsableLine) {
      strcpy (curLine, "");
      return (PARSEFILE_EOF);
   }

   // Remove \n or \r at end of line
   if (strpbrk(curLine, "\n\r")) {
      *(strpbrk(curLine, "\n\r")) = '\0';
   }

   return (PARSEFILE_NOERROR);
}

// ******* Unknown edges file parser ********
